/*----------------------------------------------------------------------
  File    : train.h
  Contents: train set management (realizations of linear point process)
  Author  : Christian Borgelt
  History : 2013.02.01 file created
            2013.02.02 file reading function tns_read() added
            2013.02.19 functions trn_filter(), trn_isect() etc. added
            2013.03.07 direction parameter added to sorting functions
            2013.11.07 train reading capability made optional
            2013.12.02 surrogate generation functions added
            2013.02.04 bugs in trn_sort() and trn_reverse() fixed
            2013.03.04 data type of train length changed to SUPP
            2014.05.09 function tns_blkperm() added (blocked perms.)
            2014.10.24 data type of train length changed to TID
            2015.08.07 function trn_add() added
            2016.10.12 special train/train array functions for sequences
----------------------------------------------------------------------*/
#ifndef __TRAIN__
#define __TRAIN__
#include <math.h>
#include "tract.h"
#ifdef TRN_SURR
#include "random.h"
#endif
#ifdef TRN_READ
#include "tabread.h"
#endif
#ifdef TRN_WRITE
#include "tabwrite.h"
#endif

#ifndef INFINITY
#include <float.h>
#define INFINITY   (DBL_MAX+DBL_MAX)
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- train read/write modes --- */
#define TRN_ITEMTRN  0x00       /* train with preceding item */
#define TRN_PURETRN  0x01       /* pure train (no preceding item) */
#define TRN_ITEMPT   0x02       /* pairs of item and point */
#define TRN_PTITEM   0x03       /* pairs of point and item */
#define TRN_SORTED   0x10       /* write data sorted by points */

/* --- train sentinel --- */
#define TRN_END    INFINITY     /* sentinel for number list */

/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */

#define E_NOITEMS   (-15)       /* no frequent items found */
#define E_ITEMEXP   (-16)       /* item expected */
#define E_DUPITEM   (-18)       /* duplicate item */
#define E_NUMEXP    (-20)       /* floating point number expected */
#define E_FLDCNT    (-22)       /* too many fields */
/* gaps in error codes to keep them aligned with those in tract.h */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- a train (lpp realization) --- */
  ITEM     item;                /* associated item */
  TID      size;                /* size of the point array */
  TID      len;                 /* current length of the train */
  double   pts[1];              /* array of points (sentinel at end) */
} TRAIN;                        /* (train/linear point process real.) */

typedef struct {                /* --- a set of trains --- */
  ITEMBASE *base;               /* underlying item base */
  TID      maxlen;              /* maximum length of a train */
  size_t   extent;              /* total number of points */
  double   min;                 /* minimum point in train set */
  double   max;                 /* maximum point in train set */
  double   width;               /* last width used for tns_maxseq() */
  ITEM     maxseq;              /* maximum possible sequence length */
  ITEM     cnt;                 /* current number of trains */
  ITEM     size;                /* size of the trains array */
  TRAIN    **trains;            /* array of trains */
  double   *pts;                /* array of all points (on demand) */
  void     *buf;                /* memory for surrogate generation */
  #ifdef TRN_READ               /* if train reading capability */
  int      err;                 /* error code (file reading) */
  TABREAD  *trd;                /* table/transaction reader */
  #endif
} TRAINSET;                     /* (set of trains) */

typedef struct {                /* --- an interval --- */
  double   beg;                 /* beginning of the interval */
  double   end;                 /* end       of the interval */
} IVAL;                         /* (interval) */

#ifdef TRN_SURR
typedef TRAIN*    TRNSURRFN (TRAIN    *src, double beg, double end,
                             RNG *rng, TRAIN *dst);
typedef TRAINSET* TNSSURRFN (TRAINSET *src, double beg, double end,
                             double delta, RNG *rng, TRAINSET *dst);
#endif
/*----------------------------------------------------------------------
  Train Functions
----------------------------------------------------------------------*/
extern TRAIN*    trn_create   (ITEM item, const double *pts, TID n);
extern void      trn_delete   (TRAIN *t);
extern TRAIN*    trn_resize   (TRAIN *t, TID n);
extern TRAIN*    trn_clone    (const TRAIN *t);
extern void      trn_copy     (TRAIN *dst, const TRAIN *src);

extern ITEM      trn_item     (const TRAIN *t);
extern double*   trn_points   (TRAIN *t);
extern TID       trn_len      (const TRAIN *t);
extern int       trn_add      (TRAIN *t, double p);
extern void      trn_sort     (TRAIN *t, int dir);
extern void      trn_reverse  (TRAIN *t);
extern int       trn_cmp      (const void *p1,
                               const void *p2, void *data);
extern int       trn_cmpsz    (const void *p1,
                               const void *p2, void *data);

extern TID       trn_filter   (TRAIN *tns, double beg, double end);
extern TID       trn_filterx  (TRAIN *dst, TRAIN *src, IVAL *ivals);
extern TID       trn_seqfiltx (TRAIN *dst, TRAIN *src, IVAL *ivals,
                               double width, int clomax);
extern TID       trn_isect    (IVAL *dst, IVAL *src, TRAIN *train,
                               double width);
extern TID       trn_seqisect (IVAL *dst, IVAL *src, TRAIN *train,
                               double width);
#ifndef NDEBUG
extern void      trn_show     (TRAIN *t, ITEMBASE *base);
#endif

/*----------------------------------------------------------------------
  Surrogate Generation Functions
----------------------------------------------------------------------*/
#ifdef TRN_SURR
extern TRAIN*    trn_ident    (TRAIN *src, double beg, double end,
                               RNG *rng, TRAIN *dst);
extern TRAIN*    trn_random   (TRAIN *src, double beg, double end,
                               RNG *rng, TRAIN *dst);
extern TRAIN*    trn_dither   (TRAIN *src, double beg, double end,
                               RNG *rng, TRAIN *dst);
extern TRAIN*    trn_shift    (TRAIN *src, double beg, double end,
                               RNG *rng, TRAIN *dst);
#endif
/*----------------------------------------------------------------------
  Train Array Functions
----------------------------------------------------------------------*/
extern size_t    tna_dstsize  (ITEM trncnt, size_t ptcnt);
extern ITEM      tna_filter   (TRAIN **dst, TRAIN **src, ITEM n,
                               IVAL *ivals, SUPP supp, void **mem);
extern ITEM      tna_filterx  (TRAIN **dst, TRAIN **src, ITEM n,
                               IVAL *ivals, SUPP supp, ITEM *used,
                               void **mem);
extern ITEM      tna_seqfilt  (TRAIN **dst, TRAIN **src, ITEM n,
                               IVAL *ivals, double width, SUPP supp,
                               void **mem);
extern ITEM      tna_seqfiltx (TRAIN **dst, TRAIN **src, ITEM n,
                               IVAL *ivals, double width, SUPP supp,
                               ITEM *used, void **mem);
extern ITEM      tna_reduce   (TRAIN **dst, TRAIN **src, ITEM n,
                               IVAL *ivals, SUPP supp, void **mem);
extern ITEM      tna_seqred   (TRAIN **dst, TRAIN **src, ITEM n,
                               IVAL *ivals, double width, SUPP supp,
                               int clomax, void **mem);
extern ITEM      tna_seqredx  (TRAIN **dst, TRAIN **src, ITEM n,
                               IVAL *ivals, double width, SUPP supp,
                               ITEM *used, TRAIN **trains, void **mem);

/*----------------------------------------------------------------------
  Train Set Functions
----------------------------------------------------------------------*/
extern TRAINSET* tns_create   (ITEMBASE *base);
extern void      tns_delete   (TRAINSET *tns, int delib);
extern ITEMBASE* tns_base     (TRAINSET *tns);
extern TRAINSET* tns_clone    (TRAINSET *tns);
extern TRAINSET* tns_copy     (TRAINSET *dst, const TRAINSET *src);

extern ITEM      tns_itemcnt  (TRAINSET *tns);
extern ITEM      tns_cnt      (const TRAINSET *tns);
extern TID       tns_maxlen   (const TRAINSET *tns);
extern size_t    tns_extent   (const TRAINSET *tns);
extern double    tns_min      (TRAINSET *tns, int recalc);
extern double    tns_max      (TRAINSET *tns, int recalc);
extern ITEM      tns_maxseq   (TRAINSET *tns, double width);

extern int       tns_add      (TRAINSET *tns, TRAIN *t);
extern TRAIN*    tns_train    (TRAINSET *tns, ITEM item);
extern void      tns_sort     (TRAINSET *tns, int dir);
extern void      tns_filter   (TRAINSET *tns, double beg, double end);
extern int       tns_recode   (TRAINSET *tns, int dir);
extern double*   tns_points   (TRAINSET *tns);

#ifdef TRN_READ
extern int       tns_read     (TRAINSET *tns, TABREAD *trd, int mode);
extern CCHAR*    tns_errmsg   (TRAINSET *tns, char *buf, size_t size);
#endif
#ifdef TRN_WRITE
extern int       tns_write    (TRAINSET *tns, TABWRITE *twr,
                               int mode, const char *fmt);
#endif
#ifndef NDEBUG
extern void      tns_show     (TRAINSET *tns);
#endif

/*----------------------------------------------------------------------
  Surrogate Generation Functions
----------------------------------------------------------------------*/
#ifdef TRN_SURR
extern TRAINSET* tns_ident    (TRAINSET *src, double beg, double end,
                               double delta, RNG *rng, TRAINSET *dst);
extern TRAINSET* tns_random   (TRAINSET *src, double beg, double end,
                               double delta, RNG *rng, TRAINSET *dst);
extern TRAINSET* tns_dither   (TRAINSET *src, double beg, double end,
                               double delta, RNG *rng, TRAINSET *dst);
extern TRAINSET* tns_shift    (TRAINSET *src, double beg, double end,
                               double delta, RNG *rng, TRAINSET *dst);
extern TRAINSET* tns_kernest  (TRAINSET *src, double beg, double end,
                               double delta, RNG *rng, TRAINSET *dst);
extern TRAINSET* tns_permute  (TRAINSET *src, double beg, double end,
                               double delta, RNG *rng, TRAINSET *dst);
extern TRAINSET* tns_permblk  (TRAINSET *src, double beg, double end,
                               double delta, RNG *rng, TRAINSET *dst);
#endif
/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define trn_delete(t)     free(t)
#define trn_item(t)       ((t)->item)
#define trn_points(t)     ((const double*)(t)->pts)
#define trn_len(t)        ((t)->len)
#define trn_sort(t,d)     dbl_qsort  ((t)->pts, (size_t)(t)->len, d)
#define trn_reverse(t)    dbl_reverse((t)->pts, (size_t)(t)->len)

/*--------------------------------------------------------------------*/
#define tna_dstsize(n,p)  ((size_t)(n)*sizeof(TRAIN) \
                          +((size_t)n+p)*sizeof(double))

/*--------------------------------------------------------------------*/
#define tns_base(t)       ((t)->base)
#define tns_itemcnt(t)    ib_cnt((t)->base)
#define tns_cnt(t)        ((t)->cnt)
#define tns_maxlen(t)     ((t)->maxlen)
#define tns_extent(t)     ((t)->extent)
#define tns_train(t,i)    ((t)->trains[i])

#endif
